"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.main = void 0;
const typescript_1 = __importDefault(require("typescript"));
const config_1 = require("./util/config");
const fs_1 = require("./util/fs");
const path_1 = require("./util/path");
const project_1 = require("./util/project");
const runner_1 = require("./runner");
const errors_1 = require("./util/errors");
const debug_1 = require("./util/debug");
const progress_1 = require("./progress");
const main = async (options) => {
    const { cwd, workingDir, configFilePath: configFilePathArg, tsConfigFilePath: tsConfigFilePathArg, include, exclude, ignore, gitignore, isIncludeEntryFiles, isDev, isShowProgress, jsDoc, debug, } = options;
    const updateMessage = (0, progress_1.getMessageUpdater)(options);
    (0, debug_1.debugLogObject)(options, 1, 'Unresolved configuration', options);
    updateMessage('Reading configuration and manifest files...');
    const manifestPath = await (0, fs_1.findFile)(cwd, workingDir, 'package.json');
    const manifest = manifestPath && require(manifestPath);
    const configFilePath = configFilePathArg ?? 'knip.json';
    const resolvedConfigFilePath = await (0, fs_1.findFile)(cwd, workingDir, configFilePath);
    const localConfig = resolvedConfigFilePath && require(resolvedConfigFilePath);
    if (configFilePathArg && !resolvedConfigFilePath) {
        throw new errors_1.ConfigurationError(`Unable to find ${configFilePathArg}`);
    }
    const tsConfigFilePath = tsConfigFilePathArg ?? 'tsconfig.json';
    const resolvedTsConfigFilePath = await (0, fs_1.findFile)(cwd, workingDir, tsConfigFilePath);
    if (tsConfigFilePathArg && !resolvedTsConfigFilePath) {
        throw new errors_1.ConfigurationError(`Unable to find ${tsConfigFilePathArg}`);
    }
    let tsConfigPaths = [];
    if (resolvedTsConfigFilePath) {
        const config = typescript_1.default.readConfigFile(resolvedTsConfigFilePath, typescript_1.default.sys.readFile);
        tsConfigPaths = config.config.compilerOptions?.paths
            ? Object.keys(config.config.compilerOptions.paths).map(p => p.replace(/\*/g, '**'))
            : [];
        if (config.error) {
            throw new errors_1.ConfigurationError(`Unable to read ${(0, path_1.relative)(resolvedTsConfigFilePath)}`);
        }
    }
    const dir = (0, path_1.relative)(workingDir);
    const resolvedConfig = (0, config_1.resolveConfig)(manifest.knip ?? localConfig, { workingDir: dir, isDev });
    (0, debug_1.debugLogObject)(options, 1, 'Resolved configuration', resolvedConfig);
    if (!resolvedConfigFilePath && !manifest.knip && !resolvedTsConfigFilePath) {
        throw new errors_1.ConfigurationError(`Unable to find ${configFilePath} or package.json#knip or ${tsConfigFilePath}`);
    }
    const { entryFiles, productionFiles, projectFiles } = await (async () => {
        if (resolvedConfig) {
            const skipAddFiles = { skipAddingFilesFromTsConfig: true, skipFileDependencyResolution: true };
            const projectOptions = resolvedTsConfigFilePath
                ? { tsConfigFilePath: resolvedTsConfigFilePath }
                : { compilerOptions: { allowJs: true } };
            updateMessage('Resolving entry files...');
            const entryPaths = await (0, path_1.resolvePaths)({
                workingDir,
                patterns: resolvedConfig.entryFiles,
                ignore,
                gitignore,
            });
            (0, debug_1.debugLogFiles)(options, 1, 'Globbed entry paths', entryPaths);
            const production = (0, project_1.createProject)({ ...projectOptions, ...skipAddFiles }, entryPaths);
            const entryFiles = production.getSourceFiles();
            (0, debug_1.debugLogSourceFiles)(options, 1, 'Included entry source files', entryFiles);
            production.resolveSourceFileDependencies();
            const productionFiles = production.getSourceFiles();
            (0, debug_1.debugLogSourceFiles)(options, 1, 'Included production source files', productionFiles);
            updateMessage('Resolving project files...');
            const projectPaths = await (0, path_1.resolvePaths)({
                workingDir,
                patterns: resolvedConfig.projectFiles,
                ignore,
                gitignore,
            });
            (0, debug_1.debugLogFiles)(options, 1, 'Globbed project paths', projectPaths);
            const project = (0, project_1.createProject)({ ...projectOptions, ...skipAddFiles }, projectPaths);
            const projectFiles = project.getSourceFiles();
            (0, debug_1.debugLogSourceFiles)(options, 1, 'Included project source files', projectFiles);
            return { entryFiles, productionFiles, projectFiles };
        }
        else {
            updateMessage('Resolving project files...');
            const project = (0, project_1.createProject)({ tsConfigFilePath: resolvedTsConfigFilePath });
            const files = project.getSourceFiles();
            return { entryFiles: files, productionFiles: files, projectFiles: files };
        }
    })();
    const report = (0, config_1.resolveIncludedIssueTypes)(include, resolvedConfig ? exclude : ['files'], resolvedConfig);
    const config = {
        workingDir,
        report,
        entryFiles,
        productionFiles,
        projectFiles,
        isIncludeEntryFiles: !resolvedConfig || isIncludeEntryFiles,
        dependencies: Object.keys(manifest.dependencies ?? {}),
        peerDependencies: Object.keys(manifest.peerDependencies ?? {}),
        optionalDependencies: Object.keys(manifest.optionalDependencies ?? {}),
        devDependencies: Object.keys(manifest.devDependencies ?? {}),
        isDev: Boolean(resolvedConfig?.dev),
        tsConfigPaths,
        isShowProgress,
        jsDocOptions: {
            isReadPublicTag: jsDoc.includes('public'),
        },
        debug,
    };
    const { issues, counters } = await (0, runner_1.findIssues)(config);
    (0, debug_1.debugLogObject)(options, 2, 'Issues', issues);
    return { report, issues, counters };
};
exports.main = main;
